use anyhow::Result;
use std::num::NonZeroU8;

use super::{MaximizerAgent, MinimizerAgent, RandomAgent};
use crate::{
	agents::util::free_slots_by_influence,
	board::{Board, PlayerAction},
	util::Symb,
};

pub struct SimpleMinimax {}

impl SimpleMinimax {
	fn step(&mut self, board: &Board, minimize: bool) -> Result<PlayerAction> {
		let available_numbers = (0..=9)
			.map(|x| match x {
				0 => Symb::Zero,
				x => Symb::Number(NonZeroU8::new(x).unwrap()),
			})
			.filter(|x| !board.contains(*x))
			.collect::<Vec<_>>();

		// For the code below, we must guarantee that
		// min_slots + max_slots <= available_numbers.len
		let n_free = board.get_board().iter().filter(|x| x.is_none()).count();
		if available_numbers.len() < n_free || n_free >= 10 {
			return RandomAgent {}.step_min(board);
		}

		let t = free_slots_by_influence(&board);

		if t.len() == 0 {
			return RandomAgent {}.step_min(board);
		}

		let (pos, val) = t[0];

		let symb = {
			if minimize {
				if val >= 0.0 {
					available_numbers[0]
				} else {
					available_numbers[available_numbers.len() - 1]
				}
			} else {
				if val <= 0.0 {
					available_numbers[0]
				} else {
					available_numbers[available_numbers.len() - 1]
				}
			}
		};

		Ok(PlayerAction { symb, pos })
	}
}

impl MinimizerAgent for SimpleMinimax {
	fn step_min(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, true)
	}
}

impl MaximizerAgent for SimpleMinimax {
	fn step_max(&mut self, board: &Board) -> Result<PlayerAction> {
		self.step(board, false)
	}
}
